
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"

#include "../../src/initng_control_command.h"

#include "initng_ngc2.h"

static void accepted_client(void);
static void closesock(void);
static void doread(FILE * file_sock);
static void addsock(void);
static int sendping(void);
static int open_socket(void);
static void check_socket(void);

/*
   In the Linux implementation, sockets which are visible in the filesystem
   honour the permissions of the directory they are in.  Their owner, group
   and their permissions can be changed.  Creation of a new socket will
   fail if the process does not have write and search (execute) permission
   on the directory the socket is created in.  Connecting to the socket
   object requires read/write permission.  This behavior differs from  many
   BSD derived systems which ignore permissions for Unix sockets.  Portable
   programs should not rely on this feature for security.
 */


/* globals */
int serversock;
struct stat sock_stat;
const char *socket_filename;

static void closesock(void)
{
    /* Check if we need to remove hooks */
    if (serversock == 0)
        return;
    D_("closesock %d\n", serversock);
    initng_del_hook(FDWATCHERS, &accepted_client);

    /* close socket and set to 0 */
    close(serversock);
    serversock = 0;
}

static void doread(FILE * file_sock)
{
    char buf[512];              /* the READ buffer */
    int len = 0;                /* read buffer string lenght */
    char *arg = NULL;

    s_command *tmp_cmd;         /* temprary storage for a command */

    assert(file_sock);

    D_("Client accepted, reading from socket\n");

    /* use file descriptor, because fread hangs here? */
    len = read(file_sock->_fileno, buf, sizeof(buf));
    if (len < 1)
    {
        D_("Bad, did not get anything! len=%i\n", len);
        return;
    }

    buf[len] = '\0';
    /* ping check : */
    if (buf[0] == 'X')
    {
        D_("Ping received, sending pong\n");
        fprintf(file_sock, "Y");
        return;
    }

    tmp_cmd = initng_command_find_by_command_id(buf[0]);
    if (!tmp_cmd)
    {
        F_("Invalid command type '%c', string : %s", buf[0], buf);
        fprintf(file_sock, "E");
        return;
    }

    if (buf[1] == '\0')
        arg = NULL;
    else
        arg = buf + 1;

    switch (tmp_cmd->com_type)
    {

        case TRUE_OR_FALSE_COMMAND:
        case INT_COMMAND:
            {
                int ret = 0;

                assert(tmp_cmd->u.int_command_call);
                D_("Calling an int command.\n");

                ret = (int) (*tmp_cmd->u.int_command_call) (arg);
                fprintf(file_sock, "%i", ret);
                break;
            }
        case STRING_COMMAND:
            {
                char *send_buf = NULL;

                assert(tmp_cmd->u.string_command_call);
                D_("Calling an string command.\n");

                send_buf = (*tmp_cmd->u.string_command_call) (arg);
                if (send_buf)
                {
                    fprintf(file_sock, "%s", send_buf);
                    free(send_buf);
                }
                else
                    fprintf(file_sock, "\n");
                break;
            }
        case PRINT_COMMAND:
            {
                assert(tmp_cmd->u.print_command_call);
                D_("Calling an print command.\n");

                (*tmp_cmd->u.print_command_call) (arg, file_sock);
                break;
            }
        case VOID_COMMAND:
            {
                assert(tmp_cmd->u.void_command_call);
                D_("Calling an void command!\n");

                (*tmp_cmd->u.void_command_call) (arg);
                fprintf(file_sock, "\n");
                break;
            }

        default:
            fprintf(file_sock, "E");
            F_("Invalid command type '%c', string : %s", buf[0], buf);
            return;
    }
    D_("Returned successfully.\n");
    return;
}

/* this adds the serversock to the global select */
static void addsock(void)
{
    D_("addsock %d\n", serversock);
    if (serversock >= 3)
        initng_add_hook(FDWATCHERS, 30, &accepted_client, serversock);
    else
        F_("libngc2 trying to add a fdhook that is not >= 3\n");
}

/* called by fd hook, when data is no socket */
void accepted_client(void)
{
    int newsock;
    FILE *file_sock;

    D_("Got here from fd hook.\n");
    /* we try to fix socket after every service start
       if it fails here chances are a user screwed it
       up, and we shouldn't manually try to fix anything. */
    if (serversock <= 0)
    {
        /* socket is invalid but we were called, call closesocket to make sure it wont happen again */
        F_("accepted client called with serversock %d, report bug\n",
           serversock);
        closesock();
        F_("Attempting to reopen socket\n");
        open_socket();
        return;
    }

    /* create a new socket, for reading */
    if ((newsock = accept(serversock, NULL, NULL)) > 0)
    {
        file_sock = fdopen(newsock, "r+");
        D_("read socket open, now reading\n");
        /* read the data, by the doread function */
        doread(file_sock);
        /* clean up, and return */
        fclose(file_sock);
        D_("Success!\n");
        return;
    }

    /* temporary unavaible */
    if ((errno == EAGAIN) || (errno == EWOULDBLOCK))
    {
        W_("errno = EAGAIN!\n");
        return;
    }

    /* This'll generally happen on shutdown, dont cry about it. */
    D_("Error accepting socket %d, %s\n", serversock, strerror(errno));
    closesock();
    return;
}

/* Send a ping to ourselfes to check if we're 100% ok. */
static int sendping()
{
    int client;
    struct sockaddr_un sockname;
    char buf[2];

    D_("Sending ping\n");

    /* Create the socket. */
    client = socket(PF_UNIX, SOCK_STREAM, 0);
    if (client < 0)
    {
        F_("Failed to init socket\n");
        return FALSE;
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;

    strcpy(sockname.sun_path, socket_filename);

    /* Try to connect */
    if (connect
        (client, (struct sockaddr *) &sockname,
         (strlen(sockname.sun_path) + sizeof(sockname.sun_family))) < 0)
    {
        close(client);
        return FALSE;
    }

    /* Write X to ping ourselfes */

    if (write(client, "X", 1) != 1)
    {
        F_("Could not send ping\n");
        close(client);
        return FALSE;
    }

    /* Accept "server side" */
    accepted_client();

    /* Note, this could hardlock, but I dont see how it can in any way since the write is ok. */
    if ((read(client, buf, 2) != 1) && buf[0] != 'Y')
    {
        F_("Could not recieve pong, buf[0] : %c\n", buf[0]);
        close(client);
        return FALSE;
    }
    close(client);

    D_("Got pong\n");
    return TRUE;
}

/* this will try to open a new socket */
static int open_socket()
{
    /*    int flags; */
    struct sockaddr_un serv_sockname;

    D_("Creating " SOCKET_ROOTPATH " dir\n");

    /* Make /dev/initng if it doesn't exist (try either way) */
    if (mkdir(SOCKET_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1
        && errno != EEXIST)
    {
        F_("Coud not create " SOCKET_ROOTPATH "\n");
        return (FALSE);
    }

    /* chmod root path for root use only */
    if (chmod(SOCKET_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1)
    {
        /* path doesn't exist, we dont have /dev yet. */
        if (errno == ENOENT || errno == EROFS)
            return (FALSE);

        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           SOCKET_ROOTPATH);
    }

    closesock();

    /* Create the socket. */
    serversock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (serversock < 1)
    {
        F_("Failed to init socket (%s)\n", strerror(errno));
        serversock = 0;
        return (FALSE);
    }

    /* Set socket to non blocking mode */
    /*    flags = fcntl(serversock, F_GETFL);
       if (flags < 0)
       {
       F_("Failed to fcntl serversock\n");
       closesock();
       return (FALSE);
       }
       flags |= O_NONBLOCK;
       if (fcntl(serversock, F_SETFL, flags) < 0)
       {
       F_("Failed to set serversock O_NONBLOCK\n");
       closesock();
       return (FALSE);
       } */


    /* Bind a name to the socket. */
    serv_sockname.sun_family = AF_UNIX;

    strcpy(serv_sockname.sun_path, socket_filename);

    /* Remove old socket file if any */
    unlink(serv_sockname.sun_path);

    /* Try to bind */
    if (bind
        (serversock, (struct sockaddr *) &serv_sockname,
         (strlen(serv_sockname.sun_path) +
          sizeof(serv_sockname.sun_family))) < 0)
    {
        F_("Error binding to socket (errno: %d str: '%s')\n", errno,
           strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }

    /* chmod socket for root only use */
    if (chmod(serv_sockname.sun_path, S_IRUSR | S_IWUSR) == -1)
    {
        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           serv_sockname.sun_path);
        closesock();
        return (FALSE);
    }

    /* store sock_stat for checking if we need to recreate socket later */
    stat(serv_sockname.sun_path, &sock_stat);

    /* Listen to socket */
    if (listen(serversock, 5))
    {
        F_("Error on listen (errno: %d str: '%s')\n", errno, strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }


    /* att socket to initng select, with hook */
    addsock();

    /* Run check : */
    if (!sendping())
    {
        F_("Sendping check failed, ngc2 comunication not available (if you see this open a bug)\n");
        closesock();
        return (FALSE);
    }

    return (TRUE);
}

/* this will check socket, and reopen on failure */
static void check_socket()
{
    struct stat st;

    D_("Checking socket\n");

    /* Check if socket needs reopening */
    if (serversock <= 0)
    {
        D_("serversock not set, opening new socket.\n");
        open_socket();
        return;
    }

    /* stat the socket, reopen on failure */
    memset(&st, 0, sizeof(st));
    if (stat(socket_filename, &st) < 0)
    {
        W_("Stat failed! Opening new socket.\n");
        open_socket();
        return;
    }

    /* compare socket file, with the one that we know, reopen on failure */
    if (st.st_dev != sock_stat.st_dev || st.st_ino != sock_stat.st_ino
        || st.st_mtime != sock_stat.st_mtime)
    {
        F_("Invalid socket found, reopening\n");
        open_socket();
        return;
    }

    D_("Socket ok.\n");
    return;
}

/* this function, will make a check for socket, on every new service that goes up */
int service_status(active_h * service)
{
    /* only try open, when a service got up */
    if (service->a_status == DONE)
        check_socket();

    return (TRUE);
}

int module_init(void)
{
    D_("module_init(ngc2);\n");

    /* zero globals */
    serversock = 0;
    memset(&sock_stat, 0, sizeof(sock_stat));


    /* decide which socket to use */
    if (g.i_am_init)
        socket_filename = SOCKET_FILENAME_REAL;
    else
        socket_filename = SOCKET_FILENAME_TEST;
    D_("Socket is: %s\n", socket_filename);

    D_("adding hook, that will reopen socket, for every started service.\n");
    initng_add_hook(WATCHERS, 50, &service_status);

    /* do the first socket directly */
    open_socket();

    D_("ngc2.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(ngc2);\n");

    /* close open sockets */
    closesock();

    /* remove hooks */
    initng_del_hook(WATCHERS, &service_status);

    D_("ngc2.so.0.0 unloaded!\n");

}
